#include <iostream>

#include "../include/Config.h"
#include "../include/CovertChannel/Channels/TCPAppendChannel.hpp"
#include "../include/CovertChannel/Channels/TCPOptionCustomChannel.hpp"
#include "../include/CovertChannel/Channels/TCPOptionTimestampChannel.hpp"
#include "../include/CovertChannel/Channels/TCPUrgencyChannel.hpp"
#include "../include/CovertChannel/CovertChannel.h"
#include "../include/CovertChannel/ForwardChannel.h"
#include "../include/Queue.h"
#include "../include/Server.h"
#include "../include/UserManager.h"

using namespace std;

int main(int argc, char *argv[]) {
	// load config int namespace
	if (!Config::init("config.txt")) {
		cerr << "configuration could not be loaded properly" << endl;
		exit(EXIT_FAILURE);
	}

	CovertChannel *covertchannel = nullptr;
	const std::string covertChannelMode = Config::getValue("covertChannelMode");
	if (covertChannelMode == "tcpurgency") {
		const string innerInterface = Config::getValue("innerInterface");
		const string outerInterface = Config::getValue("outerInterface");

		const string ownIP = Config::getValue("ownIP");
		const string targetIP = Config::getValue("targetIP");
		const string targetPort = Config::getValue("targetPort");
		const string passiveMode = Config::getValue("passiveMode");

		if (passiveMode == "true") {
			covertchannel = new TCPUrgencyChannel<true>(innerInterface, outerInterface, ownIP, targetIP, targetPort);
		} else {
			covertchannel = new TCPUrgencyChannel<false>(innerInterface, outerInterface, ownIP, targetIP, targetPort);
		}

		// covertchannel = new ForwardChannel(innerInterface, outerInterface);
		covertchannel->startSniffing();
	} else if (covertChannelMode == "tcpappend") {
		const string innerInterface = Config::getValue("innerInterface");
		const string outerInterface = Config::getValue("outerInterface");

		const string ownIP = Config::getValue("ownIP");
		const string targetIP = Config::getValue("targetIP");
		const string targetPort = Config::getValue("targetPort");
		const string passiveMode = Config::getValue("passiveMode");

		if (passiveMode == "true") {
			covertchannel = new TCPAppendChannel<8, true>(innerInterface, outerInterface, ownIP, targetIP, targetPort);
		} else {
			covertchannel = new TCPAppendChannel<8, false>(innerInterface, outerInterface, ownIP, targetIP, targetPort);
		}

		// covertchannel = new ForwardChannel(innerInterface, outerInterface);
		covertchannel->startSniffing();
	} else if (covertChannelMode == "tcpoptiontimestamp") {
		const string innerInterface = Config::getValue("innerInterface");
		const string outerInterface = Config::getValue("outerInterface");

		const string ownIP = Config::getValue("ownIP");
		const string targetIP = Config::getValue("targetIP");
		const string targetPort = Config::getValue("targetPort");
		const string passiveMode = Config::getValue("passiveMode");

		if (passiveMode == "true") {
			covertchannel = new TCPOptionTimestampChannel<true>(innerInterface, outerInterface, ownIP, targetIP, targetPort);
		} else {
			covertchannel = new TCPOptionTimestampChannel<false>(innerInterface, outerInterface, ownIP, targetIP, targetPort);
		}

		// covertchannel = new ForwardChannel(innerInterface, outerInterface);
		covertchannel->startSniffing();
	} else if (covertChannelMode == "tcpoptioncustom") {
		const string innerInterface = Config::getValue("innerInterface");
		const string outerInterface = Config::getValue("outerInterface");

		const string ownIP = Config::getValue("ownIP");
		const string targetIP = Config::getValue("targetIP");
		const string targetPort = Config::getValue("targetPort");
		const string passiveMode = Config::getValue("passiveMode");
		const string sendFile = Config::getValue("sendFile");

		if (passiveMode == "true") {
			covertchannel = new TCPOptionCustomChannel<8, true>(innerInterface, outerInterface, ownIP, targetIP, targetPort);
		} else {
			covertchannel = new TCPOptionCustomChannel<8, false>(innerInterface, outerInterface, ownIP, targetIP, targetPort);
		}

		// covertchannel = new ForwardChannel(innerInterface, outerInterface);
		covertchannel->startSniffing();
	} else if (covertChannelMode == "forward") {
		const string innerInterface = Config::getValue("innerInterface");
		const string outerInterface = Config::getValue("outerInterface");

		covertchannel = new ForwardChannel(innerInterface, outerInterface);
		covertchannel->startSniffing();
	}

	// check if userStorage is add specified location
	// if not create one
	UserManager::init(Config::getValue("userdatabase"));

	// Init queue so the file transfers can be executed
	Queue::channel = covertchannel;

	try {
		io_service io_service;
		Server server(io_service);
		io_service.run();
	} catch (exception &e) {
		cerr << e.what() << endl;
	}

	if (covertchannel == nullptr) {
		delete (covertchannel);
	}

	return 0;
}
